import time
import json

from loguru import logger
from langchain_core.prompts import ChatPromptTemplate

from nl2query.core.base_module import BaseModule
from nl2query.intent_engine.schema import IntentJsonSchema
from nl2query.core.llm_models import get_llm
from nl2query.intent_engine.prompts import get_intent_prompt


class IntentEngine(BaseModule):
    """Concrete implementation of BaseModule for intent detection"""

    def __init__(
        self,
        system_prompt: str,
        pydantic_class: IntentJsonSchema = IntentJsonSchema,
        metadata: str = None,
        examples: str = None,
        config: str = None,
        intent_structure: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            system_prompt=system_prompt,
            pydantic_class=pydantic_class,
            examples=examples,
            *args,
            **kwargs,
        )

        self.prompt = system_prompt
        self.intent_structure = intent_structure
        self.config = config
        self.pydantic_class = pydantic_class

        logger.info(f"Initialized IntentEngine with prompt: {system_prompt}")

    def run(self, state):
        """Process the state and return intent JSON"""
        try:
            start_time = time.time()
            self.metadata = state["metadata"]  # TODO remove
            model_type = state.get("model_type", "openai")
            model_name = state.get("model_name", "gpt-4o")
            temperature = state.get("temperature", 0.01)

            if state["query_reframer_yn"]:
                self.query = state["reframed_query"]
            else:
                self.query = state["query"]

            prompt = get_intent_prompt(
                self.prompt, self.examples, self.metadata, self.intent_structure
            )

            prompt = ChatPromptTemplate.from_messages(
                [("system", prompt), ("human", "{query}")]
            )

            llm = get_llm(
                model_type=model_type,
                model_name=model_name,
                temperature=temperature,
            )
            structured_llm = llm.with_structured_output(self.pydantic_class)
            few_shot_structured_llm = prompt | structured_llm

            response = few_shot_structured_llm.invoke({"query": self.query})
            intent_json = response.dict()["intent_json"]
            logger.info(f"Intent JSON: {intent_json}")
            if "error_message" in intent_json:
                err = intent_json["error_message"]
                raise Exception(f"Intent Engine Error: {err}")

            if isinstance(intent_json, str):
                intent_json_rbac = json.loads(intent_json)
            elif isinstance(intent_json, dict):
                intent_json_rbac = intent_json
            else:
                raise TypeError(f"Unexpected type for intent_json: {type(intent_json)}")

            operation = intent_json_rbac.get("operation")
            if operation != "read":
                raise Exception(
                    f"Intent Engine Error: Operation {operation} not supported"
                )
            else:
                state["intent_json"] = intent_json
                logger.info(f"Intent JSON: {intent_json}")
                state["raw_messages"].append(
                    {"role": "intent_json", "content": response}
                )
                end_time = time.time()
                response_time = end_time - start_time
                state["intent_engine_processing_time"] = response_time

                logger.info(
                    f"Intent engine processing time: {response_time:.4f} seconds"
                )

                return state, state["intent_json"]

        except json.JSONDecodeError as e:
            logger.error(f"Failed to parse intent JSON: {e}")
            return {"error": "Invalid JSON response from LLM"}
        except Exception as e:
            logger.error(f"Error processing intent: {e}")
            raise



from nl2query.agent.state import State
from nl2query.core.decorators import wrap_state_transition
from loguru import logger


@wrap_state_transition
def select_next_node_after_tables_selector(state: State):
    """Select next node after tables selector."""
    return "query_reframer_node" if state["query_reframer_yn"] else "intent_engine_node"


@wrap_state_transition
def select_next_node_after_query_reframer(state: State):
    """Select next node after query reframer."""
    return "intent_engine_node" if state["intent_yn"] else "query_builder_node"


@wrap_state_transition
def select_next_node_after_intent_engine(state: State):
    """Select next node after intent engine."""
    if state.get("proceed_to_query_builder_yn"):
        return "query_builder_node"
    if state["intent_ambiguity_yn"]:
        return "intent_ambiguity_checker_node"
    if state["intent_filterable_yn"]:
        return "intent_filter_checker_node"
    return "query_builder_node"


@wrap_state_transition
def select_next_node_after_intent_ambiguity_checker_engine(state: State):
    """Select next node after intent engine."""
    if state["intent_ambiguity"].get("exists"):
        return "intent_ambiguity_handler_node"
    else:
        if state["intent_filterable_yn"]:
            return "intent_filter_checker_node"
        else:
            logger.info("Intent ambiguity does not exist.")
            if state["regenerate_intent_yn"]:
                return "intent_engine_node"
            else:
                return "query_builder_node"


@wrap_state_transition
def select_next_node_after_intent_filter_checker(state: State):
    """Select next node after intent filter checker."""
    return "query_builder_node"


@wrap_state_transition
def select_next_node_after_query_builder(state: State):
    """Select next node after query builder."""
    return "query_corrector_node"


@wrap_state_transition
def select_next_node_after_query_corrector(state: State):
    """Select next node after query corrector."""
    if state["query_executor_yn"]:
        return "query_executor_node"
    else:
        return "user_followup_query_node"


@wrap_state_transition
def select_next_node_after_query_executor(state: State):
    """Select next node after query executor"""
    return "user_followup_query_node"


@wrap_state_transition
def select_next_node_after_followup_query(state: State):
    if state["tables_selector_yn"]:
        return "tables_selector_node"
    elif state["query_reframer_yn"]:
        return "query_reframer_node"
    elif state["intent_yn"]:
        return "intent_engine_node"
    else:
        return "query_builder_node"



from loguru import logger

from nl2query.core.decorators import handle_exceptions


class StartNode:
    def __init__(self, config):
        self._config = config

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at initial node.")
            state["state_id"] = 0
            state["conversation_messages"] = []
            state["errors"] = []
            state["query"] = state["query"]
            if not state["query"]:
                raise Exception("Error at StartNode: 'query'")

            state["user_message"] = []
            state["raw_messages"] = []
            state["messages"] = []
            state["intent_ambiguity"] = {"ambiguous_fields": {}}
            state["intent_filters"] = {"filterable_fields": {}}
            state["regenerate_intent_yn"] = False
            state.update(self._config)
            return state
        except Exception as e:
            raise Exception(f"Error at StartNode: {e}")


class TableSelectorNode:
    def __init__(self, table_selector):
        self._table_selector = table_selector

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at tables selector node.")
            state["state_id"] = 1
            _, response = self._table_selector.run(state)
            state["selected_tables"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at TableSelectorNode: {e}")


class QueryReframerNode:
    def __init__(self, query_reframer):
        self._query_reframer = query_reframer

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at query reframer node.")
            state["state_id"] = 2
            _, response = self._query_reframer.run(state)
            state["reframed_query"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at QueryReframerNode: {e}")


class IntentEngineNode:
    def __init__(self, intent_engine):
        self._intent_engine = intent_engine

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at intent engine node.")
            state["state_id"] = 3
            _, response = self._intent_engine.run(state)
            state["intent_json"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at IntentEngineNode: {e}")


class IntentAmbiguityCheckerNode:
    def __init__(self, intent_ambiguity):
        self._intent_ambiguity = intent_ambiguity

    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at intent ambiguity checker node.")
            state["state_id"] = 4
            state = self._intent_ambiguity.process_ambiguity(state)
            return state
        except Exception as e:
            raise Exception(f"Error at IntentAmbiguityHandlerNode: {e}")


class IntentAmbiguityHandlerNode:
    def __init__(self, intent_ambiguity):
        self._intent_ambiguity = intent_ambiguity

    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at intent ambiguity handler.")
            state["state_id"] = 5
            state = self._intent_ambiguity.handle_ambiguity(state)
            # logger.info(response)
            return state
        except Exception as e:
            raise Exception(f"Error at IntentAmbiguityHandlerNode: {e}")


class IntentFilterCheckerNode:
    def __init__(self, intent_filter):
        self._intent_filter = intent_filter

    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at intent filter handler.")
            state["state_id"] = 6
            return state
        except Exception as e:
            raise Exception(f"Error at IntentFilterCheckerNode: {e}")


class QueryBuilderNode:
    def __init__(self, query_builder):
        self._query_builder = query_builder

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at query builder node.")
            state["state_id"] = 7
            state, response = self._query_builder.run(state)
            state["initial_query"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at QueryBuilderNode: {e}")


class QueryCorrectorNode:
    def __init__(self, query_corrector):
        self._query_corrector = query_corrector

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at query corrector node.")
            state["state_id"] = 8
            _, response = self._query_corrector.run(state)
            state["validated_query"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at QueryCorrectorNode: {e}")


class QueryExecutorNode:
    def __init__(self, query_executor):
        self._query_executor = query_executor

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at query executor node.")
            state["state_id"] = 9
            _, response = self._query_executor.run(state)
            state["output_response"] = response
            return state
        except Exception as e:
            raise Exception(f"Error at QueryExecutorNode: {e}")


class UserFollowUpNode:
    def __init__(self):
        pass  # TODO add here

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            # logger.info("I am here at User followup node.")
            state["state_id"] = 10
            state["intent_ambiguity"] = {"ambiguous_fields": {}}
            state["intent_filters"] = {"filterable_fields": {}}
            state["proceed_to_query_builder_yn"] = False
            return state
        except Exception as e:
            raise Exception(f"Error at UserFollowUpNode: {e}")


class LastNode:
    def __init__(self):
        pass  # TODO add here

    @handle_exceptions
    def __call__(self, state, **kwargs):
        try:
            logger.info("I am here at last node.")
            state["state_id"] = -1
            return state
        except Exception as e:
            raise Exception(f"Error at LastNode: {e}")



import json
import time
from loguru import logger
from langchain_core.prompts import ChatPromptTemplate

from nl2query.core.base_module import BaseModule
from nl2query.table_selector.schema import TableSelectorSchema
from nl2query.core.llm_models import get_llm
from nl2query.table_selector.prompts import get_table_selector_prompts


class TableSelector(BaseModule):
    """Concrete implementation of BaseModule for intent detection"""

    def __init__(
        self,
        system_prompt: str = None,
        pydantic_class: TableSelectorSchema = TableSelectorSchema,
        prompt: str = None,
        examples: str = None,
        table_info: str = None,
        metadata_file_path: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            system_prompt=system_prompt,
            pydantic_class=pydantic_class,
            examples=examples,
            *args,
            **kwargs,
        )

        self.prompt = prompt
        self.examples = examples
        self.pydantic_class = pydantic_class
        self.table_info = table_info
        self.metadata_file_path = metadata_file_path  # Store metadata file path

        logger.info(f"Initialized IntentEngine with prompt: {system_prompt}")

    def load_metadata(self):
        """Load the metadata either from the given JSON file path or directly from a JSON string."""
        try:
            if isinstance(
                self.metadata_file_path, str
            ) and self.metadata_file_path.startswith("{"):
                # If the metadata_file_path is a JSON string
                logger.info("Directly using provided JSON metadata.")
                return json.loads(self.metadata_file_path)

            elif isinstance(self.metadata_file_path, str):
                # If it's a file path, load the metadata from the file
                logger.info(f"Loading metadata from file: {self.metadata_file_path}")
                with open(self.metadata_file_path, "r") as file:
                    metadata = json.load(file)
                return metadata

            else:
                logger.warning(
                    "Invalid metadata source provided. Expected a file path or a JSON string."
                )
                return {}

        except Exception as e:
            logger.error(f"Error loading metadata: {e}")
            raise

    def get_metadata_for_selected_tables(self, selected_tables):
        """Select the corresponding metadata from the JSON based on selected tables."""
        if not selected_tables:
            logger.info("No tables selected, skipping metadata retrieval.")
            return {}

        metadata = self.load_metadata()
        selected_metadata = {}

        for table in selected_tables:
            if table is None:
                logger.debug("Encountered null in selected_tables, skipping.")
                continue

            if table in metadata:
                selected_metadata[table] = metadata[table]
            else:
                logger.warning(f"Metadata for table {table} not found.")

        return selected_metadata

    def escape_metadata_braces(self, metadata):
        """Recursively replace '{' with '{{' and '}' with '}}' in the metadata."""
        if isinstance(metadata, dict):
            return {
                key: self.escape_metadata_braces(value)
                for key, value in metadata.items()
            }
        elif isinstance(metadata, str):
            return metadata.replace("{", "{{").replace("}", "}}")
        else:
            return metadata

    def run(self, state):
        """Process the state and return intent JSON"""
        try:
            start_time = time.time()
            self.model_type = state.get("model_type", "openai")
            self.model_name = state.get("model_name", "gpt-4o")
            self.temperature = state.get("temperature", 0.01)
            self.query = state["query"]

            prompt = get_table_selector_prompts(self.prompt, self.table_info)

            prompt = ChatPromptTemplate.from_messages(
                [("system", prompt), ("human", "{query}")]
            )

            llm = get_llm(
                model_type=self.model_type,
                model_name=self.model_name,
                temperature=self.temperature,
            )
            structured_llm = llm.with_structured_output(self.pydantic_class)
            few_shot_structured_llm = prompt | structured_llm

            response = few_shot_structured_llm.invoke({"query": self.query})

            selected_table = response.dict()["tables"]
            selected_table_reason = response.dict()["selected_reason"]

            logger.info(f"Selected tables:{selected_table}")
            logger.info(f"Selected reason:{selected_table_reason}")
            state["selected_tables"] = selected_table
            state["selected_table_reason"] = selected_table_reason

            selected_metadata = self.get_metadata_for_selected_tables(selected_table)
            logger.info(f"Selected metadata: {selected_metadata}")

            # metadata = json.loads(selected_metadata)

            # escaped_metadata_str = metadata_str.replace("{", "{{").replace("}", "}}")

            state["metadata"] = selected_metadata

            state["raw_messages"].append(
                {"role": "table_selector", "content": response}
            )
            formatted_message = {"role": "user", "content": state["query"]}
            state["messages"].append(formatted_message)
            end_time = time.time()
            response_time = end_time - start_time
            state["table_selector_response_time"] = response_time
            logger.info(f"Table selector processing time: {response_time}")

            return state, selected_table

        except Exception as e:
            logger.error(f"Error processing intent: {e}")
            raise



import json
from loguru import logger

from nl2query.core.path_setup import input_dir
from nl2query.core.base_module import BaseModule


class IntentAmbiguityHandler(BaseModule):
    """
    A class to handle ambiguity in intent fields.
    The run() method processes ambiguity (if any) and, if the user has provided input,
    it handles that input to update how the query is reframed.
    """

    def __init__(
        self,
        system_prompt: str = None,
        examples: str = None,
        ambiguous_fields_file: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(system_prompt, examples, *args, **kwargs)
        if ambiguous_fields_file is None:
            ambiguous_fields_file = f"{input_dir}/ambigious_fields.json"
        with open(ambiguous_fields_file, "r") as f:
            self.ambiguous_data = json.load(f)

    def process_ambiguity(self, state: dict) -> dict:
        """
        Checks if the intent in state contains any ambiguous fields.
        If found, marks them as unresolved, picks the first for processing,
        and generates a clarification message which is set in state["user_message"].
        Otherwise, it marks ambiguity as not existing.
        """
        ambiguous_data_mapping = self.ambiguous_data.get("mapping", {})
        intent_ambiguity_data = state.get("intent_ambiguity")

        if "exists" not in intent_ambiguity_data:
            ambiguity_check = False
            intent_entities = [
                field
                for entity in state["intent_json"].get("entities", [])
                for field in entity.get("fields")
            ]

            for entity in intent_entities:
                if entity in ambiguous_data_mapping:
                    ambiguity_check = True
                    intent_ambiguity_data["exists"] = True
                    intent_ambiguity_data["ambiguous_fields"][
                        ambiguous_data_mapping[entity]
                    ] = "unresolved"

            if ambiguity_check:
                intent_ambiguity_data["is_processing"] = list(
                    intent_ambiguity_data["ambiguous_fields"].keys()
                )[0]
                user_message = self.generate_ambiguity_message(
                    intent_ambiguity_data["is_processing"]
                )
                state["user_message"] = user_message

            else:
                intent_ambiguity_data["exists"] = False

        else:
            if intent_ambiguity_data["exists"]:
                user_message = self.generate_ambiguity_message(
                    intent_ambiguity_data["is_processing"]
                )
                state["user_message"] = user_message

        if not intent_ambiguity_data["exists"] and not state["intent_filterable_yn"]:
            state["proceed_to_query_builder_yn"] = True

        state["intent_ambiguity"] = intent_ambiguity_data

        logger.info(f"Intent ambiguity data: {intent_ambiguity_data}")
        return state

    def generate_ambiguity_message(self, field: str) -> str:
        """
        Generates a clarification message by listing all possible options
        for the given ambiguity category from the ambiguous_fields.json.
        """
        messages = ["Did you mean:"]
        for desc in self.ambiguous_data.get("mapping_description", {}).get(field, []):
            messages.append(f"{desc.get('idx')}. {desc.get('value')}")
        return "\n".join(messages)

    def handle_ambiguity(self, state: dict) -> dict:
        """
        Processes the user’s response (stored in state["query"]) for disambiguation.
        It updates the state's reframed query with clarification details and marks the ambiguous field as resolved.
        If additional ambiguous fields remain unresolved, it updates state["user_message"] accordingly.
        """
        user_input = state.get("query", "").strip().lower()
        ambiguous_data_description = self.ambiguous_data.get("mapping_description", {})
        intent_ambiguity_data = state.get("intent_ambiguity", {})
        processing_field = intent_ambiguity_data.get("is_processing")

        # List available option indices for the current ambiguous field.
        available_options = [
            str(desc.get("idx"))
            for desc in ambiguous_data_description.get(processing_field, [])
        ]

        if user_input in available_options:
            for desc in ambiguous_data_description.get(processing_field, []):
                if str(desc.get("idx")) == user_input:
                    message = f"\nWe are interpreting the field as `{desc.get('field')}` ({desc.get('info')}).\n"
                    state["reframed_query"] = state.get("reframed_query", "") + message
                    state["regenerate_intent_yn"] = True

            logger.info(
                f"Reframed query after handling ambiguity: {state['reframed_query']}"
            )
            # Mark the current ambiguous field as resolved, only if the user_input is in available_options
            if processing_field in intent_ambiguity_data.get("ambiguous_fields", {}):
                intent_ambiguity_data["ambiguous_fields"][processing_field] = "resolved"

        # If user_input is not in available options, do not mark processing_field as resolved.

        ambiguous_fields = intent_ambiguity_data.get("ambiguous_fields", {})

        if isinstance(ambiguous_fields, dict):
            unresolved_fields = [
                field
                for field, status in ambiguous_fields.items()
                if status == "unresolved"
            ]
        else:
            logger.warning(
                f"Expected a dictionary for ambiguous_fields, but got: {type(ambiguous_fields)}"
            )
            unresolved_fields = []

        if unresolved_fields:
            intent_ambiguity_data["is_processing"] = unresolved_fields[0]
            state["user_message"] = self.generate_ambiguity_message(
                unresolved_fields[0]
            )
        else:
            intent_ambiguity_data["exists"] = False

        state["intent_ambiguity"] = intent_ambiguity_data

        logger.info(f"After handling intent ambiguity: {intent_ambiguity_data}")
        return state

    def run(self, state: dict) -> dict:
        pass



QUERY_BUILDER_PROMPT = (
    "You are an expert SQL query generator. Your task is to convert a natural language query or an intent JSON "
    "into a syntactically correct SQL query based on the provided schema. "
    "Make sure to prepend the schema name to the table names, i.e., format as SCHEMA_NAME.TABLE_NAME."
)

EXAMPLE_WITH_INTENT_JSON = """
Example:
Natural language query: "What’s the total sales for 2023?"
intent_JSON: {{"intent": "aggregate", "entities": [{{"field": "total", "value": null}}, {{"field": "date", "value": "2023"}}], "operation": "sum"}}
SQL Query: "SELECT SUM(total) FROM SALES_SCHEMA.SALES WHERE date = '2023' LIMIT 5"
"""

EXAMPLE_WITHOUT_INTENT_JSON = """
Example:
Natural language query: "What’s the total sales for 2023?"
SQL Query: "SELECT SUM(total) FROM SALES_SCHEMA.SALES WHERE date = '2023' LIMIT 5"
"""

GUIDELINES = """
Guidelines:
- Limit results to 5 unless specified otherwise.
- Return only the query as plain text, no additional text or JSON.
- Include the schema prefix (e.g., SCHEMA_NAME.TABLE_NAME) in the SQL query.
"""


def get_query_builder_prompt(prompt=None, examples=None, schema_name=None):
    """Generate a prompt for building an SQL query based on the state."""
    prompt_parts = []
    base_prompt = prompt if prompt else QUERY_BUILDER_PROMPT
    prompt_parts.append(base_prompt)

    examples_text = "\n".join(examples) if examples else EXAMPLE_WITHOUT_INTENT_JSON
    prompt_parts.append(f"Examples:\n{examples_text}")

    if schema_name:
        prompt_parts.append(
            f"Schema name provided: {schema_name}. Make sure to use the schema prefix for all table names."
        )
    if not prompt:
        prompt_parts.append(GUIDELINES)

    return "\n\n".join(prompt_parts)



import json


def load_json_file(file_path):
    with open(file_path, "r") as file:
        return json.load(file)


def load_txt_file(file_path):
    with open(file_path, "r") as file:
        return file.read()



import logging
from pprint import pformat
import sys
from loguru import logger
from loguru._defaults import LOGURU_FORMAT


class InterceptHandler(logging.Handler):
    def emit(self, record: logging.LogRecord):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def format_record(record: dict) -> str:
    format_string = LOGURU_FORMAT
    if record["extra"].get("payload") is not None:
        record["extra"]["payload"] = pformat(
            record["extra"]["payload"], indent=4, compact=True, width=88
        )
        format_string += "\n<level>{extra[payload]}</level>"
    format_string += "{exception}\n"
    return format_string


def init_logging(filename: str = None):
    """
    Replaces logging handlers with a handler for using the custom handler.

    Args:
        filename (str, optional): Base filename for log files. If provided, logs will be written to
            files like `<filename>.info.log`, `<filename>.debug.log`, etc. If None, defaults to
            'logs/info.log', 'logs/debug.log', etc.
    """
    # Remove handlers from all uvicorn loggers
    loggers = (
        logging.getLogger(name)
        for name in logging.root.manager.loggerDict
        if name.startswith("uvicorn.")
    )
    for uvicorn_logger in loggers:
        uvicorn_logger.handlers = []

    intercept_handler = InterceptHandler()
    logging.getLogger("uvicorn").handlers = [intercept_handler]

    uvicorn_access = logging.getLogger("uvicorn.access")
    uvicorn_access.propagate = True

    # Determine log file paths based on filename parameter
    base_path = filename if filename else "logs"
    info_file = f"logs/{base_path}.info.log" if filename else "logs/info.log"
    debug_file = f"logs/{base_path}.debug.log" if filename else "logs/debug.log"
    error_file = f"logs/{base_path}.error.log" if filename else "logs/error.log"

    logger.remove()
    logger.add(
        info_file,
        level="INFO",
        format=format_record,
        filter=lambda record: record["level"].name == "INFO",
    )
    logger.add(
        debug_file,
        level="DEBUG",
        format=format_record,
        filter=lambda record: record["level"].name == "DEBUG",
    )
    logger.add(
        error_file,
        level="ERROR",
        format=format_record,
        filter=lambda record: record["level"].name == "ERROR",
    )
    logger.add(sys.stdout, level="DEBUG", format=format_record, colorize=True)




# nl2query/core/models.py

import os

from langchain_openai import ChatOpenAI
from langchain_anthropic import ChatAnthropic
from langchain_google_vertexai import ChatVertexAI


class LLMModelLoader:
    """
    A generic loader for language models using LangChain.

    Parameters:
        model_type (str): Type of the model ("openai", "gemini", "claude" or "anthropic").  
                          If not provided, will be loaded from the `MODEL_TYPE` environment variable (default: "openai").
        model_name (str): The name or id of the model. If not provided, will be loaded from the `MODEL_NAME` environment variable.
        temperature (float): The sampling temperature. If not provided, will be loaded from the `MODEL_TEMPERATURE` environment variable.
        **kwargs: Any additional parameters that the model supports.
    """

    def __init__(self, model_type: str = None, model_name: str = None, temperature: float = None, **kwargs):
        self.model_type = (model_type or os.environ.get("MODEL_TYPE", "openai")).lower()
        self.model_name = model_name or os.environ.get("MODEL_NAME", "text-davinci-003")
        # Convert the temperature to float if it comes from environment
        self.temperature = temperature if temperature is not None else float(os.environ.get("MODEL_TEMPERATURE", 0.7))
        self.kwargs = kwargs

    def load_model(self):
        """
        Instantiates and returns the LLM based on the specified model_type.
        """
        if self.model_type == "openai":
            return ChatOpenAI(model_name=self.model_name, temperature=self.temperature, **self.kwargs)
        elif self.model_type in ["claude", "anthropic"]:
            # Note: The Anthropic class might expect the model name as "model" instead of "model_name".
            return ChatAnthropic(model=self.model_name, temperature=self.temperature, **self.kwargs)
        elif self.model_type == "gemini":
            return ChatVertexAI(model=self.model_name, temperature=self.temperature, **self.kwargs)
        else:
            raise ValueError(f"Unsupported model type: {self.model_type}")

def get_llm(model_type: str = None, model_name: str = None, temperature: float = None, **kwargs):
    """
    Convenience function to load and return an LLM instance.

    Usage:
        llm = get_llm(model_type="openai", model_name="gpt-3.5-turbo", temperature=0.5)
    """
    loader = LLMModelLoader(model_type, model_name, temperature, **kwargs)
    return loader.load_model()


QUERY_REFRAMER_WITH_METADATA_PROMPT = (
    "You are an expert SQL query generator. Use the provided config mapping, metadata, and examples "
    "to reframe the user's natural language query into a precise reframed query with detailed information."
)

QUERY_REFRAMER_WITH_CONFIG_PROMPT = (
    "You are an expert SQL query generator. Use the provided config mapping and examples "
    "to reframe the user's natural language query into a precise reframed query with detailed information."
)

QUERY_REFRAMER_WITH_METADATA_EXAMPLES = """
**Examples**
Example 1:
    query: What is the average sales today?
    reframed_query: What is the average of the 'total' column (sales) in the 'sales' table for transactions occurring today, considering the 'date' column?

Example 2:
    query: Give me top customers by sales.
    reframed_query: Identify the top customer IDs from the 'sales' table, ordered by the total sum of the 'total' column (sales), including details on customer name and total sales amount.
"""

QUERY_REFRAMER_WITH_CONFIG_EXAMPLES = """
**Examples**
Example 1:
    query: What is the average sales today?
    reframed_query: What is the average of the 'total' column (sales) in 'sales' table for transactions occurring today, considering the 'date' column?

Example 2:
    query: Give me top customers by sales.
    reframed_query: Identify the top customer IDs from the 'sales' table, ordered by the total sum of the 'total' column (sales), including details on customer name and total sales amount.
"""

# New config mapping prompt and examples for SQL sales queries
CONFIG_MAPPING_PROMPT_SALES = """You are an expert in mapping natural language terms to SQL database fields. Given the user's query, map the key terms or technical jargons to their corresponding table columns or conditions based on the provided config mapping. Return the mapping as a JSON object.
"""

CONFIG = """
      "technical_jargon": {{
        "customers": [
          "client details",
          "user information",
          "account holder data"
        ],
        "orders.status = 'pending'": [
          "unprocessed orders",
          "awaiting fulfillment",
          "incomplete transactions"
        ],
        "order_items.quantity > 0": [
          "active line items",
          "purchased products"
        ],
        "products.stock_quantity": [
          "inventory levels",
          "available stock"
        ],
        "orders.order_date": [
          "transaction timestamp",
          "purchase date"
        ],
        "order_items.price * order_items.quantity": [
          "line item total",
          "product revenue"
        ]
      }}
"""

CONFIG_MAPPING_EXAMPLES_SALES = """
Example 1:
user_query: What is the total sales amount for this month?
mapping_output:
{{
    "total sales amount": "sales.total"
}}

Example 2:
user_query: Who are my top 5 customers by revenue last quarter?
mapping_output:
{{
    "top 5 customers": "ORDER BY SUM(sales.total) DESC LIMIT 5",
    "revenue": "sales.total"
}}
"""


def get_reframed_query_with_config_prompt(
    # query: str,
    # config: str,
    query_reframer_with_config_prompt: str = None,
    query_reframer_with_config_examples: str = None,
) -> str:
    query_reframer_prompt = []

    config_prompt = (
        query_reframer_with_config_prompt
        if query_reframer_with_config_prompt
        else QUERY_REFRAMER_WITH_CONFIG_PROMPT
    )
    query_reframer_prompt.append(config_prompt)

    if query_reframer_with_config_examples:
        query_reframer_prompt.append(
            f"Examples:\n{query_reframer_with_config_examples}"
        )
    else:
        query_reframer_prompt.append(
            f"Examples:\n{QUERY_REFRAMER_WITH_CONFIG_EXAMPLES}"
        )
    # escaped_config = str(config).replace("{", "{{").replace("}", "}}")
    # query_reframer_prompt.append("\n**mapping_output**:\n" + escaped_config)

    if not query_reframer_with_config_prompt:
        query_reframer_prompt.append(
            "Reframe the `query` into a precise SQL query and return it as plain text."
        )

    return "\n\n".join(query_reframer_prompt)


def get_reframed_query_with_metadata_prompt(
    query: str,
    metadata: str,
    query_reframer_with_metadata_prompt: str = None,
    query_reframer_with_metadata_examples: str = None,
) -> str:
    """Generate a prompt for reframing a query using database configuration and metadata."""
    query_reframer_prompt = []

    metadata_prompt = (
        query_reframer_with_metadata_prompt
        if query_reframer_with_metadata_prompt
        else QUERY_REFRAMER_WITH_METADATA_PROMPT
    )
    query_reframer_prompt.append(metadata_prompt)

    query_reframer_prompt.append(f"**Metadata:**\n{metadata}")

    if query_reframer_with_metadata_examples:
        query_reframer_prompt.append(
            f"Examples:\n{query_reframer_with_metadata_examples}"
        )
    else:
        query_reframer_prompt.append(
            f"Examples:\n{QUERY_REFRAMER_WITH_METADATA_EXAMPLES}"
        )

    query_reframer_prompt.append(f"Given the query: '{query}'")

    if not query_reframer_with_metadata_prompt:
        query_reframer_prompt.append(
            "Reframe the `query` into a precise SQL query and return it as plain text."
        )

    prompt = "\n\n".join(query_reframer_prompt)
    return prompt


def get_config_mapping_prompt(
    config: str,
    config_mapping_prompt: str = None,
    config_mapping_examples: str = None,
) -> str:
    """Generate a prompt for mapping query terms to SQL database fields."""
    config_mapping_prompt_list = []

    prompt = (
        config_mapping_prompt if config_mapping_prompt else CONFIG_MAPPING_PROMPT_SALES
    )
    config_mapping_prompt_list.append(prompt)
    escaped_config = str(config).replace("{", "{{").replace("}", "}}")
    config_mapping_prompt_list.append("Config mapping:\n" + escaped_config)

    if config_mapping_examples:
        config_mapping_prompt_list.append(f"**Examples:**\n{config_mapping_examples}")
    else:
        config_mapping_prompt_list.append(
            f"**Examples:**\n{CONFIG_MAPPING_EXAMPLES_SALES}"
        )

    if not config_mapping_prompt:
        config_mapping_prompt_list.append("Given the query:")
        config_mapping_prompt_list.append(
            "Map the key terms in the `query` to their corresponding SQL table columns or conditions and return the mapping as a JSON object with a key 'mapping_output'. "
            "The JSON object should have a field named 'mapping_output' that contains the key-value pairs."
        )
    prompt = "\n\n".join(config_mapping_prompt_list)
    return prompt



import yaml
import uuid
from loguru import logger
from typing import Any, Optional, Dict, Set, Tuple
from IPython.display import Image, display
from langgraph.graph import START, END, StateGraph
from langgraph.checkpoint.memory import MemorySaver
from langgraph.types import Command

from nl2query.core.logger import init_logging
from nl2query.agent.state import State
from nl2query.agent.nodes import (
    StartNode,
    TableSelectorNode,
    QueryReframerNode,
    IntentEngineNode,
    QueryBuilderNode,
    QueryCorrectorNode,
    QueryExecutorNode,
    IntentAmbiguityHandlerNode,
    IntentAmbiguityCheckerNode,
    IntentFilterCheckerNode,
    LastNode,
    UserFollowUpNode,
)
from nl2query.agent.node_selector import (
    select_next_node_after_tables_selector,
    select_next_node_after_query_reframer,
    select_next_node_after_intent_engine,
    select_next_node_after_intent_ambiguity_checker_engine,
    select_next_node_after_intent_filter_checker,
    select_next_node_after_query_builder,
    select_next_node_after_query_corrector,
    select_next_node_after_query_executor,
    select_next_node_after_followup_query,
)


class NL2QueryGraph:
    def __init__(
        self,
        config_file_path: str = "config.yaml",
        table_selector_instance=None,
        query_reframer_instance=None,
        intent_engine_instance=None,
        query_builder_instance=None,
        query_validator_instance=None,
        query_executor_instance=None,
        intent_ambiguity_instance=None,
        intent_filter_checker=None,
        interrupt_before=None,
        log_file_name: str = "logs",
    ):
        with open(config_file_path, "r") as file:
            self.config = yaml.safe_load(file)
        logger.info(f"Config loaded: {self.config}")

        self.query_reframer = query_reframer_instance
        self.table_selector = table_selector_instance
        self.intent_engine = intent_engine_instance
        self.query_builder = query_builder_instance
        self.query_validator = query_validator_instance
        self.query_executor = query_executor_instance
        self.intent_ambiguity_handler = intent_ambiguity_instance
        self.intent_filter_checker = intent_filter_checker
        self.interrupt_before = interrupt_before

        self.graph = StateGraph(State)
        self.memory = MemorySaver()
        self.initial_state = State()
        self.edges: Set[Tuple[str, str]] = set()
        self.conditional_edges: Dict[str, Dict[str, Any]] = {}

        self._setup_nodes()
        self._setup_edges()
        self._apply_edges()
        self.compiled_graph = self.compile()
        init_logging(filename=log_file_name)

    def _setup_nodes(self):
        self.graph.add_node("start_node", StartNode(config=self.config))
        logger.info("Added node: start_node")
        if self.config.get("tables_selector_yn", False):
            self.graph.add_node(
                "tables_selector_node",
                TableSelectorNode(table_selector=self.table_selector),
            )
            logger.info("Added node: tables_selector_node")
        if self.config.get("query_reframer_yn", False):
            self.graph.add_node(
                "query_reframer_node",
                QueryReframerNode(query_reframer=self.query_reframer),
            )
            logger.info("Added node: query_reframer_node")
        if self.config.get("intent_yn", False):
            self.graph.add_node(
                "intent_engine_node", IntentEngineNode(intent_engine=self.intent_engine)
            )
            logger.info("Added node: intent_engine_node")
        if self.config.get("intent_ambiguity_yn", False):
            self.graph.add_node(
                "intent_ambiguity_checker_node",
                IntentAmbiguityCheckerNode(
                    intent_ambiguity=self.intent_ambiguity_handler
                ),
            )
            logger.info("Added node: intent_ambiguity_checker_node")
        if self.config.get("intent_ambiguity_yn", False):
            self.graph.add_node(
                "intent_ambiguity_handler_node",
                IntentAmbiguityHandlerNode(
                    intent_ambiguity=self.intent_ambiguity_handler
                ),
            )
            logger.info("Added node: intent_ambiguity_handler_node")
        if self.config.get("intent_filterable_yn", False):
            self.graph.add_node(
                "intent_filter_checker_node",
                IntentFilterCheckerNode(intent_filter=self.intent_filter_checker),
            )
            logger.info("Added node: intent_filter_checker_node")
        if self.config.get("query_builder_yn", False):
            self.graph.add_node(
                "query_builder_node", QueryBuilderNode(query_builder=self.query_builder)
            )
            logger.info("Added node: query_builder_node")
        if self.config.get("query_correcter_yn", False):
            self.graph.add_node(
                "query_corrector_node",
                QueryCorrectorNode(query_corrector=self.query_validator),
            )
            logger.info("Added node: query_corrector_node")
        if self.config.get("query_executor_yn", False):
            self.graph.add_node(
                "query_executor_node",
                QueryExecutorNode(query_executor=self.query_executor),
            )
            logger.info("Added node: query_executor_node")
        self.graph.add_node("user_followup_query_node", UserFollowUpNode())
        self.graph.add_node("last_node", LastNode())
        logger.info("Added node: last_node")

    def _get_default_edge_mapping(self) -> Dict[str, Dict[str, Any]]:
        return {
            "start_node": {
                "condition": lambda state: "tables_selector_node"
                if state.get("tables_selector_yn", False)
                else "query_reframer_node",
                "targets": {"tables_selector_node", "query_reframer_node"},
            },
            "tables_selector_node": {
                "condition": select_next_node_after_tables_selector,
                "targets": {"query_reframer_node", "intent_engine_node", "last_node"},
            },
            "query_reframer_node": {
                "condition": select_next_node_after_query_reframer,
                "targets": {"intent_engine_node", "query_builder_node", "last_node"},
            },
            "intent_engine_node": {
                "condition": select_next_node_after_intent_engine,
                "targets": {
                    "intent_ambiguity_checker_node",
                    "intent_filter_checker_node",
                    "query_builder_node",
                    "last_node",
                },
            },
            "intent_ambiguity_checker_node": {
                "condition": select_next_node_after_intent_ambiguity_checker_engine,
                "targets": {
                    "intent_ambiguity_handler_node",
                    "query_builder_node",
                    "intent_engine_node",
                    "intent_filter_checker_node",
                    "last_node",
                },
            },
            "intent_filter_checker_node": {
                "condition": select_next_node_after_intent_filter_checker,
                "targets": {"query_builder_node", "intent_engine_node", "last_node"},
            },
            "query_builder_node": {
                "condition": select_next_node_after_query_builder,
                "targets": {"query_corrector_node", "last_node"},
            },
            "query_corrector_node": {
                "condition": select_next_node_after_query_corrector,
                "targets": {
                    "user_followup_query_node",
                    "query_executor_node",
                    "last_node",
                },
            },
            "query_executor_node": {
                "condition": select_next_node_after_query_executor,
                "targets": {
                    "table_selector_node",
                    "user_followup_query_node",
                    "last_node",
                },
            },
            "user_followup_query_node": {
                "condition": select_next_node_after_followup_query,
                "targets": {
                    "tables_selector_node",
                    "query_reframer_node",
                    "intent_engine_node",
                    "query_builder_node",
                    "last_node",
                },
            },
        }

    def _setup_edges(self):
        self.edges.add((START, "start_node"))
        logger.info("Added edge to edges set: START -> start_node")
        self.conditional_edges.update(self._get_default_edge_mapping())
        self.edges.add(
            ("intent_ambiguity_handler_node", "intent_ambiguity_checker_node")
        )
        self.edges.add(("intent_filter_handler_node", "intent_filter_checker_node"))

        self.edges.add(("last_node", END))

    def _apply_edges(self):
        for from_node, to_node in self.edges:
            if from_node in self.graph.nodes or from_node == START:
                if to_node in self.graph.nodes or to_node == END:
                    self.graph.add_edge(from_node, to_node)
                    logger.info(f"Applied edge: {from_node} -> {to_node}")
                else:
                    logger.warning(
                        f"Skipping edge {from_node} -> {to_node}: Target node missing"
                    )
            else:
                logger.warning(
                    f"Skipping edge {from_node} -> {to_node}: Source node missing"
                )

        # Apply conditional edges
        for source, config_data in self.conditional_edges.items():
            if source in self.graph.nodes:
                valid_targets = config_data["targets"] & set(self.graph.nodes.keys())
                if valid_targets:
                    self.graph.add_conditional_edges(
                        source, config_data["condition"], list(valid_targets)
                    )
                    logger.info(
                        f"Applied conditional edges from {source} to {valid_targets}"
                    )
                else:
                    self.graph.add_edge(source, "last_node")
                    logger.info(
                        f"No valid targets for {source}, applied edge to last_node"
                    )
            else:
                logger.warning(f"Source node {source} not in graph, skipping")

    def remove_edge(self, from_node: str, to_node: str):
        edge = (from_node, to_node)
        if edge in self.edges:
            self.edges.remove(edge)
            logger.info(f"Removed edge: {from_node} -> {to_node}")
        elif (
            from_node in self.conditional_edges
            and to_node in self.conditional_edges[from_node]["targets"]
        ):
            self.conditional_edges[from_node]["targets"].remove(to_node)
            logger.info(f"Removed conditional target {to_node} from {from_node}")

    def add_edge(self, from_node: str, to_node: str):
        if (from_node in self.graph.nodes or from_node == START) and (
            to_node in self.graph.nodes or to_node == END
        ):
            self.edges.add((from_node, to_node))
            logger.info(f"Added edge: {from_node} -> {to_node}")
        else:
            logger.warning(
                f"Cannot add edge {from_node} -> {to_node}: One or both nodes not in graph"
            )

    def add_conditional_edge(
        self, from_node: str, condition: callable, targets: Set[str]
    ):
        if from_node in self.graph.nodes:
            self.conditional_edges[from_node] = {
                "condition": condition,
                "targets": targets,
            }
            logger.info(
                f"Added conditional edge for {from_node} with targets {targets}"
            )
        else:
            logger.warning(
                f"Cannot add conditional edge for {from_node}: Node not in graph"
            )

    def compile(self) -> Any:
        return self.graph.compile(
            checkpointer=self.memory, interrupt_before=self.interrupt_before
        )

    def visualize(self):
        compiled_graph = self.compile()
        display(Image(compiled_graph.get_graph().draw_mermaid_png()))

    def list_added_nodes(self):
        logger.info("Listing all nodes in the graph:")
        for node in self.compiled_graph.nodes:
            logger.info(f"Node: {node}")
        return list(self.graph.nodes)

    # TODO run(thread_id, {user_query, followup_query})
    def run(
        self,
        thread_id=None,
        state: Optional[State] = None,
        updated_state=None,
        *args,
        **kwargs,
    ) -> Any:
        """
        Executes the graph internally until completion.
        Extra kwargs (query, model_name, etc.) can be injected into the state if needed.
        Returns the final result instead of a generator.
        """
        thread_id = thread_id or str(uuid.uuid4())
        config = {"configurable": {"thread_id": thread_id}, "recursion_limit": 2000}
        response_type = state.get("response_type", "regular") if state else "regular"

        if response_type == "streaming":
            logger.info("Running in streaming mode with feedback loop...")
            input_state = state
            final_response = []
            # while True:
            iterator = self.compiled_graph.stream(
                input_state, config=config, stream_mode="values"
            )
            for event in iterator:
                final_response.append(str(event))
            input_state = None
            current_state = self.compiled_graph.get_state(config)
            return current_state

        elif response_type == "regular":
            input_state = state
            if updated_state:
                self.compiled_graph.invoke(
                    Command(update=updated_state, resume="Lets go"), config=config
                )
            else:
                self.compiled_graph.invoke(input_state, config=config)

            input_state = None
            current_state = self.compiled_graph.get_state(config)

            return current_state
        else:
            raise ValueError(f"Unknown response_type: {response_type}")



from setuptools import setup, find_packages

setup(
    name="mle-nl2query",
    version="1.1.20",
    packages=find_packages(),
    install_requires=[
        "requests",
    ],
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",  # Supported Python versions
        "License :: OSI Approved :: MIT License",  # License for your package
        "Operating System :: OS Independent",  # Specifies that your package works across different operating systems
    ],
    python_requires=">=3.6",
    author="Palistha Deshar",
    author_email="palisthadeshar@gmail.com",
    description="nl2query package",
    url="https://github.com/RamailoTech/mle_nl2query",
    license="MIT",  # License of your package
)



import os
from dotenv import load_dotenv

load_dotenv()


PG_USER = os.getenv("PG_USER")
PG_PASSWORD = os.getenv("PG_PASSWORD")
PG_HOST = os.getenv("PG_HOST")
PG_PORT = os.getenv("PG_PORT")
PG_DATABASE = os.getenv("PG_DATABASE")
DATABASE_URL = f"postgresql://{PG_USER}:{PG_PASSWORD}@{PG_HOST}:{PG_PORT}/{PG_DATABASE}"


SNOWFLAKE_USER = os.getenv("SNOWFLAKE_USER")
SNOWFLAKE_PASSWORD = os.getenv("SNOWFLAKE_PASSWORD")
SNOWFLAKE_ACCOUNT = os.getenv("SNOWFLAKE_ACCOUNT")
SNOWFLAKE_WAREHOUSE = os.getenv("SNOWFLAKE_WAREHOUSE")
SNOWFLAKE_DATABASE = os.getenv("SNOWFLAKE_DATABASE")
SNOWFLAKE_SCHEMA = os.getenv("SNOWFLAKE_SCHEMA")
